
import unittest

import numpy as np
import numpy.testing as nptesting

import esig

STREAM = np.array([
    [1.0, 1.0],
    [3.0, 4.0],
    [5.0, 2.0],
    [8.0, 6.0]
])

SIGNATURE = np.array([
    1.0, 7.0, 5.0, 24.5, 19.0, 16.0, 12.5
])

LOG_SIGNATURE = np.array([
    7. , 5. , 1.5
])

class ArrayTestCase(unittest.TestCase):
    RTOL = 1e-3
    ATOL = 0.0

    def assertEqual(self, actual, expected, *args, **kwargs):
        if isinstance(actual, np.ndarray) or isinstance(expected, np.ndarray):
            nptesting.assert_equal(actual, expected)
        super().assertEqual(actual, expected, *args, **kwargs)

    def assert_allclose(self, actual, desired):
        """
        Custom assert method for Numpy arrays, built using
        the Numpy testing suite assertion methods.
        """
        nptesting.assert_allclose(
            actual, desired, rtol=self.RTOL, atol=self.ATOL
        )


class TestSignatureInterface(ArrayTestCase):

    def test_calculation_width_2_depth_0_error(self):
        width = 2
        depth = 0

        with self.assertRaises(ValueError):
            sig = esig.stream2sig(STREAM, depth)

    def test_calculation_width_2_depth_1(self):
        # Pure Python implementation 
        width = 2
        depth = 1

        sig = esig.stream2sig(STREAM, depth)

        self.assert_allclose(sig, SIGNATURE[:3])

    def test_calculation_width_2_depth_2(self):
        # tosig implementation

        width = 2
        depth = 2

        sig = esig.stream2sig(STREAM, depth)
        self.assert_allclose(sig, SIGNATURE)

    def test_calculation_width_2_depth_100_error(self):
        # Depth out of bounds for width 2
        width = 2
        depth = 100

        pattern = "Legitimate depth of 2<->\\d+ for records with width 2 exceeds limit"
        with self.assertRaisesRegex(RuntimeError, pattern):
            sig = esig.stream2sig(STREAM, depth)

    def test_dtype_validation_float64_passes(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1.0, 2.0],
             [3.0, 4.0],
             [5.0, 6.0]],
            dtype=np.float64
        )

        sig = esig.stream2sig(stream, depth)
        self.assertGreater(sig.size, 0)

    def test_dtype_validation_float32_passes(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1.0, 2.0],
             [3.0, 4.0],
             [5.0, 6.0]],
            dtype=np.float32
        )

        sig = esig.stream2sig(stream, depth)
        self.assertGreater(sig.size, 0)

    def test_dtype_validation_int64_fails(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1, 2],
             [3, 4],
             [5, 6]],
            dtype=np.int64
        )

        with self.assertRaises(TypeError):
            sig = esig.stream2sig(stream, depth)
        #self.assertGreater(sig.size, 0)

    def test_computed_size_vs_size(self):

        width = 2

        for depth in range(2, 10):
            with self.subTest(depth=depth):
                sig = esig.stream2sig(STREAM, depth)
                size = esig.sigdim(width, depth)
                self.assertEqual(size, sig.size)
        
    def test_signature_keys(self):
        width = 2
        depth = 2

        keys = esig.sigkeys(width, depth)
        self.assertEqual(keys, " () (1) (2) (1,1) (1,2) (2,1) (2,2)")


class TestLogSignatureInterface(ArrayTestCase):

    def test_calculation_width_2_depth_0_error(self):
        width = 2
        depth = 0

        with self.assertRaises(ValueError):
            log_sig = esig.stream2logsig(STREAM, depth)

    def test_calculation_width_2_depth_1(self):
        # Pure Python implementation 
        width = 2
        depth = 1

        log_sig = esig.stream2logsig(STREAM, depth)

        self.assert_allclose(log_sig, LOG_SIGNATURE[:2])

    def test_calculation_width_2_depth_2(self):
        # tosig implementation

        width = 2
        depth = 2

        log_sig = esig.stream2logsig(STREAM, depth)
        self.assert_allclose(log_sig, LOG_SIGNATURE)

    def test_calculation_width_2_depth_100_error(self):
        # Depth out of bounds for width 2
        width = 2
        depth = 100

        pattern = "Legitimate depth of 2<->\\d+ for records with width 2 exceeds limit"
        with self.assertRaisesRegex(RuntimeError, pattern):
            log_sig = esig.stream2logsig(STREAM, depth)

    def test_dtype_validation_float64_passes(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1.0, 2.0],
             [3.0, 4.0],
             [5.0, 6.0]],
            dtype=np.float64
        )

        log_sig = esig.stream2logsig(stream, depth)
        self.assertGreater(log_sig.size, 0)

    def test_dtype_validation_float32_passes(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1.0, 2.0],
             [3.0, 4.0],
             [5.0, 6.0]],
            dtype=np.float32
        )

        log_sig = esig.stream2logsig(stream, depth)
        self.assertGreater(log_sig.size, 0)

    def test_dtype_validation_int64_fails(self):

        width = 2
        depth = 2

        stream = np.array(
            [[1, 2],
             [3, 4],
             [5, 6]],
            dtype=np.int64
        )

        with self.assertRaises(TypeError):
            log_sig = esig.stream2logsig(stream, depth)
        #self.assertGreater(sig.size, 0)

    def test_computed_size_vs_size(self):

        width = 2

        for depth in range(2, 10):
            with self.subTest(depth=depth):
                log_sig = esig.stream2logsig(STREAM, depth)
                size = esig.logsigdim(width, depth)
                self.assertEqual(size, log_sig.size)
        
    def test_signature_keys(self):
        width = 2
        depth = 2

        keys = esig.logsigkeys(width, depth)
        self.assertEqual(keys, " 1 2 [1,2]")


import unittest
import numpy as np
from esig import tosig as ts
from unittest import TestCase
from esig.tests import auxiliaryfunct as ax
from esig.tests import esigtests as rado
from esig.tests import recombinetests as recombine


# Tree-like equivalence of paths and equivalence class signature invariance. Non-generic paths used since pruning is generically unnecessary.

a=[[0.0,0],[1,0],[1,1],[1,0],[2,0],[3,1],[2,2],[1,1],[2,2],[1,3],[2,2],[3,3],[2,2],[3,1],[4,1],[3,1],[2,0],[2,-1],[2,0],[1,0],[1,-1],[1,0],[0,0]]

b=[[0.0,0], [1,3], [0,0], [1,5], [2,5], [1,5], [0,6], [1,5], [0,0]]

c=[[0.0,0], [1,1], [3,1], [2,1], [1,1], [2,0]]

aa=ax.fullprune(a)
bb=ax.fullprune(b)
cc=ax.fullprune(c)
a=np.array(a)
b=np.array(b)
c=np.array(c)
aa=np.array(aa)
bb=np.array(bb)
cc=np.array(cc)
sigdiffa= ts.stream2sig(a,3)-ts.stream2sig(aa,3)
sigdiffb = ts.stream2sig(b,3)-ts.stream2sig(bb,3)
sigdiffc = ts.stream2sig(c,3)-ts.stream2sig(cc,3)


y=ax.random_path(1,range(-10,11),3)

x=[[0.0,0,0]]
xx = x

xx.append(y[0])
x=np.array(xx)

path1=ax.random_path(5,range(-10,11),3)
path2=ax.random_path(5,range(-10,11),3)
#print(path1)
#print(ax.concatenate(path1,path2))

sig1=ax.tensoraslevels(ts.stream2sig(np.array(path1),2),3,2)
#print(sig1)
sig2=ax.tensoraslevels(ts.stream2sig(np.array(path2),2),3,2)
#print(sig2)
sig_concat=ax.tensoraslevels(ts.stream2sig(ax.concatenate(path1,path2),2),3,2)
chen=ax.chen(sig1,sig2)
#print(chen)
#print(sig_concat)


logsig=ax.exponentiateTensor( ax.tensoraslevels( ax.Logsigastensor( np.array(path1),3),3,3))



#c=ax.Logsigastensor(a,3)
#b=ts.stream2sig(a,3)
#d=ax.exponentiateTensor(ax.tensoraslevels(c,2,3))
#print(b)
#print(d)
#b=np.array(b)
#sigdiff= ts.stream2sig(a,2)-ts.stream2sig(b,2)
#print(sigdiff)

#if all(sigdiff)==0:
 #   print("ok")



# Test Chen's theorem using tensor multiplication
# Compute esig of linear functions by hand and see it matches

# Exponentiate logsig as a tensor and see it matches esig


#a=[1,2]
#sig1=ax.linsig(a,3)
#b=[[0.0,0], [1,2]]
#b=np.array(b)
#sig2=ts.stream2sig(np.array(b),3)
#print(sig1-sig2)


#print(ax.fullunbracket(string))
#a=ax.Lietotensor(string)
#print(a)



#print(expand)

#L=ax.Lietotensor(string)
#print(L)
#print(ax.Lietotensor('1'))

#print(ax.changeofbasis(2,3))


#print(ax.Lietotensor(string))
#tensor_product=ax.fullunbracket(string)
#copy= tensor_product[1:-1]
#print(copy)

#term=ax.spotfirstterm(copy)
#print(type(term))
#print(term)
#d=ax.distribute(term)
#print(d)
#print(type(d))
##print(copy.replace(term,d))
#copy=copy.replace(term,ax.distribute(term))
#print(copy)
#term=ax.spotfirstterm(copy)
#print(term)
#print(type(term))
#a=ax.fullunbracket(string)

#print(a)
#print(len(a))
#b=a[1:-1]
#c=ax.spotterm(b)[1][1]
#print(c)

#term='1x2x(2x1-1x2)x3'
#print(ax.distribute(term))

class TestESIG(TestCase):

    def test_equivalence(self):
        self.assertEqual(all(sigdiffa),0)
        self.assertEqual(all(sigdiffb),0)
        self.assertEqual(all(sigdiffc),0)

    def test_linsig(self):
        self.assertEqual(all(ax.linsig(y[0],4) - ts.stream2sig(x,4)),0)

    def test_logsig(self):
        self.assertEqual(all(ts.stream2sig(np.array(path1),3)[1:]- logsig),0)

    def test_chen(self):
        self.assertEqual(all(np.concatenate((np.array(chen)-np.array(sig_concat)).flatten())),0)

    def test_compare_reverse(self):
        self.assertEqual(rado.compare_reverse_test(100,[-1,0,1],2,4),1)

    def test_shuffle_test(self):
        self.assertEqual(rado.shuffle_test(100,[-1,0,1],3,3),1)

    def test_reparametrizing(self):
        self.assertEqual(rado.reparametrizing_test(10,[-1,0,1],2,3,3),1)

class TestRecombine(TestCase):

    def test_recombine(self):
        self.assertEqual(recombine.test(), 0)



from enum import Enum
import os
from os.path import expanduser
import platform
from setuptools.dist import Distribution
import sys
import textwrap


# Auxiliary classes used by setup.py.


# Tell setuptools that package is a binary distribution.
# https://lucumr.pocoo.org/2014/1/27/python-on-wheels/
class BinaryDistribution(Distribution):
    def is_pure(self):
        return False


class InstallationConfiguration(object):
    """
    Installation configuration class, providing details required to install esig.
    """
    def __init__(self, package_root):
        print("Starting esig installer...")
        self.package_root = package_root

        reported_platform = platform.system().lower()
        platform_map = {
            'windows': PLATFORM.WINDOWS,
            'linux': PLATFORM.LINUX,
            'linux2': PLATFORM.LINUX,
            'darwin': PLATFORM.MACOS,
        }

        if reported_platform not in platform_map.keys():
            raise Exception(reported_platform + " not a recognised platform.")

        self.platform = platform_map[reported_platform]
        self.is64bit = sys.maxsize > 2**32

    @property
    def include_dirs(self):
        """
        Return list of directories to search for include files.

        Returns:
            list of directory paths.
        """
        # libalgebra and recombine sources + wrapper code for Python.
        # TODO: remove dependency on "recombine" having been cloned into /build/recombine

        dirs = [
            os.path.join(".", "src"),
            os.path.join(".", "libalgebra"),
            os.path.join(".", "recombine"),
            os.path.join(".", "build", "recombine", "recombine")
        ]

        if 'BOOST_ROOT' in os.environ and os.environ['BOOST_ROOT'] != '':
            dirs.append(os.environ['BOOST_ROOT'])

        if self.platform == PLATFORM.MACOS:
            dirs.append('/opt/local/include/')
        elif self.platform == PLATFORM.LINUX:
            dirs.append('/usr/include/')

        return dirs

    @property
    def library_dirs(self):
        """
        Return a list of directories to search for libraries.

        Returns:
            list of directory paths.
        """
        if not 'BOOST_ROOT' in os.environ:
            boost_rootenv = ""
        else:
            boost_root_env = os.environ['BOOST_ROOT']

        dirs = []

        if self.platform == PLATFORM.WINDOWS:
            if self.is64bit:
                lib1, lib2 = 'lib64', 'x64'
            else:
                lib1, lib2 = 'lib32', 'win32'

            dirs.append(os.path.join(boost_root_env, lib1 + '-msvc-14.0'))
            dirs.append(os.path.join(boost_root_env, lib2, 'lib'))

            # ouch, depend on MKL having been installed by rcombine
            dirs.append(
               os.path.join(
                  os.environ['USERPROFILE'],
                  "Miniconda3",
                  "pkgs",
                  "mkl-static-2021.1.1-intel_52",
                  "Library",
                  "lib"
               )
            )

            # todo: lose hardcoded knowledge of recombine installation dir
            dirs.append(os.path.join(expanduser("~"), "lyonstech", "lib"))

        # On a Mac, our best guess for including libraries will be from /opt/local/lib.
        # This is where Macports and Homebrew installs libraries to.
        elif self.platform == PLATFORM.MACOS:
            if 'DYLD_LIBRARY_PATH' in os.environ and os.environ['DYLD_LIBRARY_PATH'] != '':
                dirs = dirs + os.environ['DYLD_LIBRARY_PATH'].split(os.pathsep)

        elif self.platform == PLATFORM.LINUX:
            if 'LD_LIBRARY_PATH' in os.environ and os.environ['LD_LIBRARY_PATH'] != '':
                dirs = dirs + os.environ['LD_LIBRARY_PATH'].split(os.pathsep)

        return dirs

	# Python extension code built with distutils is compiled with the same set of compiler options,
	# regardless of whether it's C or C++. We use C _and_ C++, which rules out certain compiler options.
    @property
    def extra_compile_args(self):
        """
        Returns list of additional platform/compiler-dependent compiler arguments.

        Returns:
            list of string arguments.
        """
        args = []

        if self.platform == PLATFORM.WINDOWS:
            args.append('/EHsc')
            args.append('/DWINVER=0x0601')
            args.append('/D_WIN32_WINNT=0x0601')
            args.append('/D_SCL_SECURE_NO_WARNINGS')
            args.append('/bigobj')
        else:
            # Clang will reject this when compiling C
            if self.platform == PLATFORM.LINUX:
                args.append('-std=c++11') # want c99 as well, but not possible (see above)
            args.append('-Wno-unused-but-set-variable') # moans on some platforms

        return args

    @property
    def linker_args(self):
        """
        Return list of additional platform/compiler-dependent linker arguments.
        """
        args = []

        # How can we statically link for MACOS/LINUX? -static does not work on Linux.
        if self.platform == PLATFORM.MACOS:
            args.append('-static')

        return args

    @property
    def esig_version(self):
        """
        Extract version number from VERSION file found in package root.

        Returns:
            string representing version number, in format MAJOR.MINOR.PATCH.
        """
        version_path = os.path.join(self.package_root, 'esig', 'VERSION')

        with open(version_path, 'r') as version_file:
            return (version_file.read().strip()).replace(' ', '.')

    @property
    def long_description(self):
        """
        Extract contents of README.md file found in package root.

        Returns:
            string contents of file.
        """
        readme_path = os.path.join(self.package_root, 'README.md')

        with open(readme_path, 'r') as f:
            return f.read()

    @property
    def used_libraries(self):
        """
        List libraries used by esig. On Windows, library selection is automatic, so Boost isn't included.

        Returns:
            list of library names
        """
        libs = {
            PLATFORM.WINDOWS: ['recombine'],
            # on the Mac, recombine is a framework, not a library; needs special treatment in setup.py
            PLATFORM.MACOS: ['boost_system-mt','boost_thread-mt'],
            PLATFORM.LINUX: ['boost_system','boost_thread', 'recombine'],
        }

        return libs[self.platform]


class PLATFORM(Enum):
    WINDOWS = 1
    LINUX = 2
    MACOS = 3



import sys
import os
import unittest
from . import test_package


def run_tests(terminate=False): # argument ignored
    """
    Acquires the ESig test suite, and runs the tests.

    Returns:
        terminates with exit code 1 if all tests pass successfully; 0 otherwise.
    """
    package_root = os.path.dirname(os.path.abspath(__file__))
    suite = unittest.TestLoader().discover(package_root)
    print(suite.countTestCases() + " test cases found.")
    runner = unittest.TextTestRunner()
    result = runner.run(suite)
    print(result)

    if not result.wasSuccessful():
        sys.exit(1)
    else:
        sys.exit(0)



import numpy as np
from esig import tosig as ts
import math
import random
import itertools

def prune (path):
    """
    Function prunes tree-like components of a path, passing a single time over the list of increments. The input needs to be very specific. In particular, linearly dependant increments need to be either equal or inverse. For exhaustive pruning, use 'fullprune'. In its use in 'fullprune' the increments have supremum norm 1.

    Args:
        path (list): path given as a list of increments.
       

    Returns:
        path (list): pruned path as a list of increment vectors. Numerical values will be of type np.float64 or np.int32

    """
    path=np.array(path)
   
    l=len(path)
    copy = path-path
    it=iter(range (l))
    array=path

    for i in it:
        if i<l-1:    
            if any(array[i] + array[i+1]): 
                copy[i] = array[i]
            else: i=next(it)
        else: copy[i] = array[i]  
    copy=np.array(copy)  
    copy=list(copy)
    copy=[list(x) for x in copy]
    copy=[x for x in copy if any(x)]
    #copy=np.array(copy)
      
    return copy


def fullprune (path, o=None, x_0=None):
    """

    Function loops 'prune' to exhaustively prune the input path. It refines the path to ensure that the use of 'prune' is made with the correct type of input. 

    Args:
        path (list): path given either as a list of position vectors or list of increments.
        o (bool): optional parameter, if 'True', x_0 is required and the input is treated as increments
        x_0 (list): position vector, specifying initial point of the path if 'o==True'

    Returns:
        path (list): pruned path as a list of position vectors. Numerical values will be of type np.float64 or np.int32

    """
    

    if x_0 and o is None: 
        raise ValueError("Set o=1!")


    if o==None:
        first=path[0]
    else:
        first=x_0


    if o is None:
        # turn path into increments
        
        l=len(path)
       
        array=np.array(path)
        incr=array-array
        for i in range(l-1):
             incr[i]=array[i+1]-array[i]

    # refine the path

    dim=len(incr[0])
    length=len(incr)
    new_length=0
    list_lengths=[]
    for i in range(length):
        new_length += np.int32(max(np.abs(incr[i])))
        list_lengths.append(np.int32(max(np.abs(incr[i]))))
    refined_path=np.zeros((new_length,dim))
    counter=0
    for i in range(length):
        for j in range(list_lengths[i]):

            refined_path[counter]=(1/ list_lengths[i])*incr[i]

            counter=counter+1

    #print(refined_path)
    incr=refined_path
    l_in=len(incr)
    #print(l_in)
    l_out=0
    while l_in > l_out:
        l_in=len(incr)
        incr=prune(incr)
        l_out=len(incr)
    #print(l_out)
    #turn increments into path
    #print(incr)
    copy=[first]
    
    if l_out > 0:     
        for i in range(1,l_out+1):
            copy.append(np.array(copy[i-1]) + np.array(incr[i-1]))

    

    copy=np.array(copy)  
    
    copy=list(copy)
    
    copy=[list(x) for x in copy]
    
    

    return copy


def linsig(position_vector, degree, start=None, end=None):
    """
    Function computes the truncated signature up to degree 'degree' of a single linear path from 0 to a position vector specified.

    Args:
        position_vector (list): a list of coordinates
        degree (int): the highest degree of a signature term to return
        start (float): optional, can input the intial time value, default 0
        stop (float): optional can input the final time value, default 1

    Returns:
        signature (np.array): vector of signature values


    """
    if start==None:
        start=0
    if end==None:
        end=1
    dimension = len(position_vector)
    keys = ts.sigkeys(dimension, degree)
    pos = keys.split()
    pos_out = []
    pos[0]='(0)'
    for row in pos:
        row = row.strip('(')
        row = row.strip(')')
        row = row.split(',')
        
        row=[int(i) for i in row]
        pos_out.append(row)
    pos_out.pop(0)
    signature=[1]
    for i in pos_out:
        sigj=1
        l=len(i)
        for j in i:
            sigj=sigj*position_vector[j-1]
        sigj=sigj*(end-start)**l
        sigj=sigj/math.factorial(l)
        signature.append(sigj)    

    signature=np.array(signature)
    return signature




def isblob(string):
    open=0
    closed =0
    for i in string:
        if i in ['(','[']:
            open=open+1
        if i in [')',']']:
            closed=closed+1
        if closed > open:
            raise SyntaxError('Not a mathematical expression!')
    if open==closed:
        t=1
    else: t=0
    return t


#split_input = input_str.strip().split(' ')

def unbracket(Lie_str):
   # Take a string starting with '[' and ending with ']' and develop it as a Lie bracket, around the central ',' character
    workon=Lie_str[1:-1]
    l=len(workon)
    for i in range(l):
        if workon[i]==',':
            if isblob(workon[:i])==0:
                 pass
            else:
                blob1=workon[:i]
                blob2=workon[i+1:]
    outstring='('+blob1+'x'+blob2+'-'+blob2+'x'+blob1+')'
            
    
   
    return outstring 


# Search for Lie brackets in a string
def spotbracket(string):
    
    l= len(string)
    firstbracket=string.find('[')
    if firstbracket==-1:
        return -1
    else:
        open=1
        closed=0
        i=firstbracket+1
        while open>closed:
        
            if string[i]=='[':
                open=open+1
            if string[i]==']':
                closed=closed+1
            i=i+1
    
        bracket = string[firstbracket:i]


    return bracket

def fullunbracket(lie_string):
    while spotbracket(lie_string)!=-1:
        a=spotbracket(lie_string)
        b=unbracket(a)
        lie_string=lie_string.replace(a,b)
        #index_a=lie_string.find(a)
        #l=len(a)
        #part1=lie_string[:index_a]
        #part2=lie_string[index_a+l:]
        #lie_string=part1+b+part2
    return lie_string

def spotroundbracket(string):
    l= len(string)
    firstbracket=string.find('(')
    if firstbracket==-1:
        return -1
    else:
        open=1
        closed=0
        i=firstbracket+1
        while open>closed:
        
            if string[i]=='(':
                open=open+1
            if string[i]==')':
                closed=closed+1
            i=i+1
    
        bracket = string[firstbracket:i]
    return bracket

def isterm(string):
    t=0
    if isblob(string)==1 and string.replace('-','') != '':
        t=1
    return t

def spotfirstterm(string,j=None):
    if j==None:
        j=0
    string=string[j:]
    l=len(string)
    for i in range(l):

        if string[i]=='-':
            if isterm(string[:i])==1:
                firstterm=string[:i]
                if firstterm.find('(')==-1:
                    return spotfirstterm(string,j=i)
                else:
                    return firstterm
    else:
        return string

def distribute(term):
    """ MUST input a valid TERM"""

    factor=spotroundbracket(term)
    if factor==-1:
        return term
    else:
        factorstring=factor[1:-1]
        #factorposition=factor[1]
        lenfac=len(factorstring)
        middle=int(math.floor(lenfac/2))
        if factorstring[middle] != '-':
            raise SyntaxError ('Not a valid Lie bracket')
        else:
            distrib = term.replace(factor,factorstring[:middle],1)+'-'+term.replace(factor,factorstring[middle+1:],1)
            return distrib

def Lietotensor(Lie_bracket):
    """
    Inputs a nested Lie bracket, returns it as a sum of the basis elements in the tensor algebra.
    """
    tensor_product=fullunbracket(Lie_bracket)
    if len(tensor_product)==1:
        return [(1, [int(tensor_product)])]
    else:
        copy= tensor_product[1:-1]
    while copy.find('(') !=-1:
        term=spotfirstterm(copy)
        d=distribute(term)
        copy=copy.replace(term,d,1)
    string=copy
    string=string.replace('x',',')
    while string.find('--') != -1:
        string=string.replace('--','+')
    while string.find('++')!=-1:
        string=string.replace('++','+')
    while string.find('+-')!=-1:
        string=string.replace('+-','-')
    string=string.replace('+',' +')
    string=string.replace('-',' -')
    string='+'+string
    string=string.split(' ')
    l=len(string)
    for i in range(l):
        j=string[i]
        s=[]
        for k in j[1:]:
            if k!=',':
                s.append(int(k))
        if j[0]=='+':
            j=(1,s)
        else:
            j=(-1,s)
        string[i]=j
    return string   

def changeofbasis(dimension, degree):
    liebasis=ts.logsigkeys(dimension,degree)
    tensorbasis=ts.sigkeys(dimension,degree)
    n=ts.logsigdim(dimension,degree)
    m=ts.sigdim(dimension, degree)
    matrix=np.zeros((n,m))
    pos = tensorbasis.split()
    columnindex = []
    pos[0]='(0)'
    for row in pos:
        row = row.strip('(')
        row = row.strip(')')
        row = row.split(',')
        
        row=[int(i) for i in row]
        columnindex.append(row)
    #print(columnindex)

    rowindex=liebasis.split()
    expand=[]
    for i in range(n):
        B=Lietotensor(rowindex[i])
        #print(B)
        for j in B:
            for k in range(m):
                #print(type(j[1]))
                #print(type(k))
                if j[1]==columnindex[k]:
                    matrix[i][k]=matrix[i][k]+j[0]
    return matrix

def Logsigastensor(array, degree):
    logsig=ts.stream2logsig(array, degree)
    dimension=len(array[0])
    M=changeofbasis(dimension,degree)
    result=np.dot(logsig,M)

    return result

def tensoraslevels(tensor, dimension, degree):
    
    sum = [np.zeros(dimension ** m) for m in range(1,degree + 1)]
    
    counter=0
    for i in range(1, degree+1):
        
        for j in range(dimension**i):
            counter=counter+1
            sum[i-1][j]=tensor[counter]
    return sum


#inputs are values in the tensor algebra given as lists of levels (from 1 to
#level), assumed 0 in level 0.
#returns their concatenation product
def multiplyTensor(a,b):
    level = len(a)
    dim = len(a[0])
    sum = [np.zeros(dim ** m) for m in range(1,level + 1)]
    for leftLevel in range(1,level):
        for rightLevel in range(1, 1 + level - leftLevel):
            sum[leftLevel + rightLevel - 1]+=np.outer(a[leftLevel - 1],b[rightLevel - 1]).flatten()
    return sum

#input is a value in the tensor algebra given as lists of levels (from 1 to
#level), assumed 0 in level 0.
#returns its exp - assumed 1 in level 0
#exp(x)-1 = x+x^2/2 +x^3/6 +x^4/24 + ...
def exponentiateTensor(a):
    out = [i.copy() for i in a]
    level = len(a)
    products = [out]
    for m in range(2,level + 1):
        t = multiplyTensor(a,products[-1])
        for j in t:
            j *= (1.0 / m)
        products.append(t)
    return np.concatenate([np.sum([p[i] for p in products],0) for i in range(level)])
    

def chen(a,b):
    level = len(a)
    dim = len(a[0])
    
    sum = [a[m] + b[m] for m in range(level)]
    for leftLevel in range(1,level):
        for rightLevel in range(1, 1 + level - leftLevel):
            sum[leftLevel + rightLevel - 1]+=np.outer(a[leftLevel - 1],b[rightLevel - 1]).flatten()
            
    return sum

def concatenate(a,b):
    la=len(a)
    dim=len(a[0])
    lb=len(b)
    aa=np.array(a)
    bb=np.array(b)
    bb=bb+aa[(la-1):la]-bb[0:1]
    #print(aa[la-1],aa[(la-1):la])
    return np.vstack([aa,bb[1:]])
   

def paths_generate(dataset_lengths, a, dims):
    """ Generate all possible paths for a chosen number of dimensions, chosen lengths of path steps and a list of allowed modifiers from one step to another
    
    Modifiers of each dimension value are selected randomly from a provided list.

    Args:
        dataset_lengths (list): path lengths for which all possible path combinations are obtained
        a (list): list of modifiers applied at each step of all paths to be obtained. A longer list means more possible paths are obtained
        dims (int): number of dimensions each path step has. Higher dimensionality means more possible paths are obtained.
     
    Returns:
        list: A list of all path combinations given the arguments provided

    Example:
        >>> print (paths_generate([3],[0,1],1))
        [[[0.0], [0.0], [0.0]], [[0.0], [0.0], [1.0]], [[0.0], [1.0], [1.0]], [[0.0], [1.0], [2.0]]]
    """
    # collecting all possible increment combinations for chosen path lengths and the range of chosen changes
    out_increments = []
    for length in dataset_lengths:
        possible_changes = list(itertools.product(a, repeat=dims)) # NUMBER OF DIMENSIONS FOR WHICH PATHS GET GENERATED CAN BE MODIFIED HERE
        temp = list(itertools.product(possible_changes,repeat=length-1))
        out_increments.extend(temp)

    # calculating paths from increments with the starting point 0
    out_paths = []
    for out in out_increments:
        out_path = [list(np.zeros(dims))]
        temp = list(np.zeros(dims))
        for element in out:
            temp = [x + y for x, y in zip(temp, element)]
            out_path.append(temp)
        out_paths.append(out_path)
    return out_paths


def random_path (l,a,dims):
    """ Generate a random path which has a chosen number of steps, given a chosen set of possible modifications and step dimensions


    Args:
        l (int): length of the random path to be produced
        a (list): list of modifiers applied at each step of all paths to be obtained
        dims (int): number of dimensions each path step has

    Returns:
        list: a random path that satisfies the constraints put into the function
    
    Example:
        >>> print (random_path(10,[1,0,-1],2))
        [[-1.0, 0.0], [-1.0, 0.0], [-2.0, 1.0], [-1.0, 2.0], [-2.0, 3.0], [-3.0, 3.0], [-2.0, 2.0], [-2.0, 3.0], [-2.0, 3.0], [-3.0, 4.0]]
    """

    x = list(itertools.product(a, repeat=dims))
    out_increments = []
    for ll in range(l):
        c = random.sample(x, 1)
        out_increments.append(c)
    out_path = []
    temp = list(np.zeros(dims))
    for element in out_increments:
        temp = [x + y for x, y in zip(temp, list(element[0]))]
        out_path.append(temp)
    return out_path



def shuffles (a_list, b_list):
    """ Generates all possible shuffles of two lists
    
    Shuffles as in when shuffling two decks of cards. The resulting shuffles only contain items in the order followed in the 2 lists put into the function. 
    E.g. lists [d,e,f] and [a,b,c] would get shuffled to, among others, form a list [a,d,b,e,c,f] but never a list [b,a,d,c,e,f]

    Args:
        a_list (list): a list of elements
        b_list (list): a list of elements

    Returns:
        list: all possible shuffles of the 2 inputted lists
    
    Example:
        >>> print (shuffles([4],[1,2,3]))
        [[4, 1, 2, 3], [1, 4, 2, 3], [1, 2, 4, 3], [1, 2, 3, 4]]

    To do:
        * this function is very inefficient for generating higher numbers of shuffles, e.g. where both lists have 10 items each

    """
    a = a_list # index values of the first list of numbers
    a_codes = []
    for i, x in enumerate(a):
        x = "a" + str(i)
        a_codes.append(x)
    b = b_list
    b_codes = [] # index values of the first list of numbers
    for i, x in enumerate(b):
        x = "b" + str(i)
        b_codes.append(x)
    c = a_codes+b_codes
    all_mutants = list(itertools.permutations(c, len(c))) # generate all possible combinations of elements from the two lists of numbers
    good_mutants = []
    for x in all_mutants: # pick out combinations suitable for inclusion in shuffled product and collect those in good_mutants list
        temp1 = []
        for v in x: # extract a_codes
            if v[0] == 'a':
                temp1.append(v)
        if temp1 == a_codes: # if order of a_code indices is as in the original, carry on
            pass
        else:
            continue

        temp2 = []
        for v in x: # extract b_codes
            if v[0] == 'b':
                temp2.append(v)
        if temp2 == b_codes: # if order of b_code indices is as in the original, add x to good_mutants
            good_mutants.append(x)

    shuffled = []
    for x in good_mutants:
        x_vals = []
        for v in x:
            if v[0] =='a':
                a_id = v[1:]
                x_vals.append(a[int(a_id)])
            else:
                b_id = v[1:]
                x_vals.append(b[int(b_id)])
        shuffled.append(x_vals)

    return shuffled




import sys
import numpy as np
from numpy.linalg import norm
import time
import esig.tosig as ts
import esig


def test ():
   failures = 0

   dimension = 240 ## big test set this to 2400 small test 240
   no_points = 100000

   print("preparing data ({},{})".format(no_points,dimension))
   tic = time.perf_counter()
   np.random.seed(0)
   data = np.random.rand(no_points,dimension)
   toc = time.perf_counter()
   print("prepared data in {n:0.4f} seconds\n".format(n=toc - tic))

   ## test 1
   tic = time.perf_counter()
   selected_points, new_weights = ts.recombine(data) ## degree = 1
   toc = time.perf_counter()
   print("Recombined {no_points} points in dimension {dimension} to {dimension2} points in {n:0.4f} seconds".format(no_points=no_points,dimension=dimension,dimension2=dimension+1,n=toc - tic))

   ## check mean preserved
   old_average=np.sum(data, 0)
   new_average=new_weights.dot(np.take(data, selected_points, 0))
   normalised_error = norm(old_average - new_average)/(norm(old_average) + norm(new_average))
   print("normalised difference in the integrals = ",normalised_error)

   ## report
   if ((len(selected_points) > dimension + 1) or (normalised_error > 1e-13)):
      failures = failures + 1
      print("test 1 failed")
   else:
      print("test 1 passed")

   ## test2
   ### the points are not spanning the full space and so the minimal set should have cardinality less than or equal rank + 1
   matrix = np.random.rand(dimension,dimension + 20)
   new_data = data.dot(matrix)
   tic = time.perf_counter()
   selected_points, new_weights = ts.recombine(new_data) ## degree = 1
   toc = time.perf_counter()
   print("\nRecombined {no_points} points in dimension {dimension} to {dimension2} points in {n:0.4f} seconds".format(no_points=no_points,dimension=dimension + 20,dimension2=dimension+1,n=toc - tic))

   ## check mean preserved
   old_average=np.sum(data, 0)
   new_average=new_weights.dot(np.take(data, selected_points, 0))
   normalised_error = norm(old_average - new_average)/(norm(old_average) + norm(new_average))
   print("normalised difference in the integrals = ",normalised_error)
   print("no of points left = ", len(selected_points))

   ## report
   if ((len(selected_points) > dimension + 1) or (normalised_error > 1e-12)):
      failures = failures + 1
      print("test 2 failed")
   else:
      print("test 2 passed")

   #test3
   ## test the degree > 1 case - match second moments
   dimension = 10
   no_points = 1000
   data = np.random.rand(no_points,dimension)

   tic = time.perf_counter()
   selected_points, new_weights = ts.recombine(data, degree = 2)
   toc = time.perf_counter()
   print("\nRecombined {no_points} points in dimension {dimension} to {m} points in {n:0.4f} seconds while preserving first and second moments".format(no_points=no_points,dimension=dimension,m=len(selected_points),n=toc - tic))

   old_average=np.sum(data, 0)
   new_average=new_weights.dot(np.take(data, selected_points, 0))
   normalised_error_in_mean = norm(old_average - new_average)/(norm(old_average) + norm(new_average))

   new_cov = np.cov(np.take(data, selected_points, 0),rowvar=False, bias=True, aweights=new_weights)
   old_cov = np.cov(data,rowvar=False, bias=True,aweights=np.full(1000,1.))
   normalised_error_in_cov = norm(old_cov - new_cov)/(norm(old_cov) + norm(new_cov))

   print ("normalised_error_in_mean {}\nnormalised_error_in_cov  {}".format(normalised_error_in_mean,normalised_error_in_cov))

   if ((normalised_error_in_mean > 1e-13) or (normalised_error_in_cov > 1e-13)):
      failures = failures + 1
      print("test 3 failed")
   else:
      print("test 3 passed")

   print ("\n")

   import platform, subprocess

   print("="*40, "System Information", "="*40)
   uname = platform.uname()
   print("System: {}".format(uname.system))
   print("Node Name: {}".format(uname.node))
   print("Release: {}".format(uname.release))
   print("Version: {}".format(uname.version))
   print("Machine: {}".format(uname.machine))
   print("Processor: {}".format(uname.processor))

   return failures



import os
from tools import install_helpers as helpers
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from tools.switch_generator import SwitchGenerator


configuration = helpers.InstallationConfiguration(os.path.dirname(os.path.realpath(__file__)))


SWITCH_GEN = SwitchGenerator()

#SWITCH_GEN = SwitchGenerator({
#    2: 5,
#    3: 5,
#    4: 2,
#    5: 2
#})


class BuildExtensionCommand(build_ext):
    """
    Extends the build_ext class, allowing for the injection of additional code into the build_ext process.
    """
    def run(self):
        """
        Attempts to import numpy and append the result of numpy.get_include() to the self.include_dirs list.
        This is to avoid the circular issue of importing numpy at the top of the module.
        See https://stackoverflow.com/a/42163080.

        Args:
            self (NumpyExtensionCommand): Instance of self.

        Returns:
            None
        """
        print("Running extra esig pre-build commands...")
        print("Building switch.h")
        SWITCH_GEN.write_file()
        print("Done")

        import numpy
        self.include_dirs.append(numpy.get_include())
        build_ext.run(self)



# https://stackoverflow.com/questions/2584595/building-a-python-module-and-linking-it-against-a-macosx-framework
if configuration.platform == helpers.PLATFORM.MACOS:
    home = os.environ["HOME"]
    os.environ['LDFLAGS'] = \
        '-F ' + home + '/lyonstech/ ' + \
        '-framework recombine ' + \
        '-Wl,-rpath,' + home + '/lyonstech/'

esig_extension = Extension(
    'esig.tosig',
    sources=[
        'src/C_tosig.c',
        'src/Cpp_ToSig.cpp',
        'src/ToSig.cpp',
        'recombine/_recombine.cpp',
        'recombine/TestVec/RdToPowers2.cpp'
    ],
    # relationship between depends and include_dirs is unclear
    depends=[
        'src/ToSig.h',
        'src/C_tosig.h',
        'src/ToSig.cpp',
        'src/switch.h',
        'recombine/TestVec/OstreamContainerOverloads.h',
        'recombine/_recombine.h'
    ],
    include_dirs=configuration.include_dirs,
    library_dirs=configuration.library_dirs,
    libraries=configuration.used_libraries,
    extra_compile_args=configuration.extra_compile_args,
    extra_link_args=configuration.linker_args,
)

package_data = {
	"esig": ["VERSION", "ERROR_MESSAGE"]
}

extras_require = {
    "iisignature-backend": ["iisignature"],
}



eager_resources = []

if configuration.platform == helpers.PLATFORM.WINDOWS:
    package_data["esig"] += [
        os.path.join("libiomp5md.dll"),
        os.path.join("recombine.dll")
    ]
    # not sure why this is needed, and if it is, why package_data also needs to mention them
    eager_resources += ["libiomp5md.dll", "recombine.dll"]


setup(
    name='esig',
    version=configuration.esig_version,

    author='Terry Lyons',
    author_email='software@lyonstech.net',
    url='http://esig.readthedocs.io/en/latest/',
    license='GPLv3',

    keywords='data streams rough paths signatures',

    description="This package provides \"rough path\" tools for analysing vector time series.",
    long_description=configuration.long_description,
    long_description_content_type="text/markdown",

    include_package_data=True,
    packages=find_packages(exclude=("tools",)),
    test_suite='esig.tests.get_suite',

    package_data=package_data,
    eager_resources=eager_resources,
    distclass=helpers.BinaryDistribution,
    ext_modules=[esig_extension],

    install_requires=['numpy>=1.7'],
    setup_requires=['numpy>=1.7'],
    tests_require=['numpy>=1.7'],
    extras_require=extras_require,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Mathematics',
        ],

    cmdclass={
        'build_ext': BuildExtensionCommand,
    }
)



# Mechanism for switching backends for computing signatures



import abc
import threading

import numpy

try:
    from esig import tosig
except ImportError:
    # Error occurs during build sequence, since tosig does not exist
    pass


try:
    import iisignature
except ImportError:
    iisignature = None

BACKENDS = {}


_BACKEND_CONTAINER = threading.local()


def get_backend():
    """
    Get the current global backend used for computing signatures.
    """
    return _BACKEND_CONTAINER.context


def set_backend(cls_or_name):
    """
    Change the backend used for computing signatures globally.
    """
    if isinstance(cls_or_name, str):
        if not cls_or_name in BACKENDS:
            raise ValueError("%s does not name a valid backend" % cls_or_name)
        _BACKEND_CONTAINER.context = BACKENDS[cls_or_name]()
    elif isinstance(cls_or_name, BackendBase):
        _BACKEND_CONTAINER.context = cls_or_name
    elif issubclass(cls_or_name, BackendBase):
        _BACKEND_CONTAINER.context = cls_or_name()
    else:
        raise TypeError("Backend must be a subclass of the BackendBase class or str")


def list_backends():
    """
    Get a list of all available backends that can be used for computing
    signatures.
    """
    return list(BACKENDS)


class BackendBase(abc.ABC):
    """
    Base class for signature/log signature computation backends.

    The required methods are `compute_signature` and `compute_log_signature`,
    which take data in the form of a Numpy array and return a Numpy array
    containing the flattened signature or log signature.
    """

    @abc.abstractmethod
    def compute_signature(self, stream, depth):
        """
        Compute the signature of the stream to required depth
        """

    @abc.abstractmethod
    def compute_log_signature(self, stream, depth):
        """
        Compute the log signature of the stream to required depth
        """

    def log_sig_dim(self, dimension, depth):
        """
        Get the number of elements in the log signature
        """
        return tosig.logsigdim(dimension, depth)

    def sig_dim(self, dimension, depth):
        """
        Get the number of elements in the signature
        """
        return tosig.sigdim(dimension, depth)

    @abc.abstractmethod
    def log_sig_keys(self, dimension, depth):
        """
        Get the keys that correspond to the elements in the log signature
        """

    @abc.abstractmethod
    def sig_keys(self, dimension, depth):
        """
        Get the keys that correspond to the elements in the signature
        """


class LibalgebraBackend(BackendBase):
    """
    Use libalgebra as a backend for computing signatures and log signatures
    of paths. This is the default option.
    """

    def __repr__(self):
        return "LibalgebraBackend"

    def compute_signature(self, stream, depth):
        return tosig.stream2sig(stream, depth)

    def compute_log_signature(self, stream, depth):
        return tosig.stream2logsig(stream, depth)

    def log_sig_keys(self, dimension, depth):
        return tosig.logsigkeys(dimension, depth)
    
    def sig_keys(self, dimension, depth):
        return tosig.sigkeys(dimension, depth)


BACKENDS["libalgebra"] = LibalgebraBackend


if iisignature:

    class IISignatureBackend(BackendBase):

        def __init__(self):
            self._log_sig_prepare_cache = {}
        
        def __repr__(self):
            return "IISignatureBackend"

        def prepare(self, dimension, depth):
            if (dimension, depth) in self._log_sig_prepare_cache:
                return self._log_sig_prepare_cache[(dimension, depth)]
            
            s = iisignature.prepare(dimension, depth)
            self._log_sig_prepare_cache[(dimension, depth)] = s
            return s

        def compute_signature(self, stream, depth):
            return numpy.concatenate([[1.0], iisignature.sig(stream, depth)], axis=0)

        def compute_log_signature(self, stream, depth):
            _, dim = stream.shape
            s = self.prepare(dim, depth)
            return iisignature.logsig(stream, s)

        def log_sig_keys(self, dimension, depth):
            s = self.prepare(dimension, depth)
            return iisignature.basis(dimension, depth)
        
        def sig_keys(self, dimension, depth):
            return tosig.sigkeys(dimension, depth)




    BACKENDS["iisignature"] = LibalgebraBackend


# set the default backend
set_backend(LibalgebraBackend)



import itertools
import unittest


import esig
from esig.backends import BackendBase




class TBackend(BackendBase):

    def compute_signature(self, stream, depth):
        pass

    def compute_log_signature(self, stream, depth):
        pass

    def log_sig_keys(self, dimension, depth):
        pass

    def sig_keys(self, dimension, depth):
        pass


class TestBackendDefaultMethods(unittest.TestCase):

    def setUp(self):
        self.backend = TBackend()
        self.sig_dim_func = lambda w, d: (w**(d+1) - 1)/(w - 1)

    def test_log_sig_dim_calc(self):
        width = 2
        depth = 2
        self.assertEqual(
            self.backend.log_sig_dim(width, depth),
            3
        )

    def test_sig_dim_calc(self):
        width = 2
        depth = 2
        self.assertEqual(
            self.backend.sig_dim(width, depth),
            self.sig_dim_func(width, depth)
        )


class TestBackendGetterMethods(unittest.TestCase):

    def setUp(self):
        esig.backends.BACKENDS["test"] = TBackend

    def tearDown(self):
        esig.set_backend(esig.backends.LibalgebraBackend)

    def test_get_default_backend(self):
        backend = esig.get_backend()

        self.assertTrue(isinstance(
            backend, 
            esig.backends.LibalgebraBackend
        ))

    def test_set_backend_by_class(self):
        backend = esig.get_backend()
        assert not isinstance(backend, TBackend)

        esig.set_backend(TBackend)
        self.assertTrue(isinstance(
            esig.get_backend(), 
            TBackend
        ))

    def test_set_backend_by_string(self):
        backend = esig.get_backend()
        assert not isinstance(backend, TBackend)

        esig.set_backend("test")
        self.assertTrue(isinstance(
            esig.get_backend(), 
            TBackend
        ))
